package org.qatools.validator;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.qatools.QATools;
import org.qatools.local.LocalDir;
import org.qatools.local.LocalDirUtil;
import org.qatools.local.LocalFile;
import org.qatools.local.LocalFileUtil;
import org.qatools.util.LongUtil;



/**
 * 
 * @author hockeyeh
 *
 */
public class DetailReport {
	private static List<Package> packages = new LinkedList<Package>();
	private static List<ErrorDetail> errorDetails = new LinkedList<ErrorDetail>();
	private static LocalDir qatoolsOrgDir = new LocalDir("target/QAToolsOrg");
	private static LocalFile notEfficientFile = new LocalFile(qatoolsOrgDir, "not-efficient.html");
	private static boolean notEfficientWarning = false;
	
	//
	// color codes
	// http://www.w3schools.com/html/html_colors.asp
	private static String GREEN = "00FF66";
	private static String RED = "FF6666";
	private static String YELLOW = "yellow";
	
	static {
		if (QATools.isDefaultReportEnabled() ||
				QATools.isDetailReportEnabled() ||
				QATools.isErrorSummaryReportEnabled()) {
			
			//
			// create QAToolsOrgDir
			LocalDirUtil.createDir(qatoolsOrgDir);
			
			//
			// create notEfficientFile
			try {
				List<String> lines = DetailReport.getPageHeader("Not Ffficient!");
				lines.add("The default report generator is not efficient. It does extra file I/O each time the validator.junitAssertNoErrors() is invoked because the APIs have no way of knowing when your test automation complets.");
				lines.add("You can force the reporting I/O to be more efficient with fewer I/O by doing the following:");
				lines.add("<ul>");
				lines.add("<li>QATools.disableDefaultReport(); - this disables the default reporting");
				lines.add("<li>DetailUtil.generateDetailReport();  - do this in your @AfterClass teardown method");
				lines.add("<li>DetailUtil.generateErrorSummaryReport();  - do this in your @AfterClass teardown method");
				lines.add("</ul>");
				lines.addAll(DetailReport.getPageFooter());
				LocalFileUtil.writeLines(notEfficientFile, lines);
			} catch (Exception e) {
				//
			}			
			
		}	//if QATools.is...
	}	//static
	
	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------
	
	public static void addDebugInfo(StackTraceElement testCallerDetails, List<String> debugs) {
		String[] parts = testCallerDetails.getClassName().split("\\.");
		//System.out.println("parts.length = " + parts.length);
		String suite = parts[parts.length-1];
		String pack = testCallerDetails.getClassName().replace("."+suite, "");
		String testCaseName = testCallerDetails.getMethodName();

		Package p = getPackage(pack);
		Suite s = getSuite(p, suite);
		TestCase tc = getTestCase(s, testCaseName);
		tc.debugs = new LinkedList<String>();
		for (String str : debugs) {
			tc.debugs.add(str);
		}
	}
	
	// -------------------------------------------------------------
	// ----- Writing files one test case at a time
	// ----- is invoked by Validator.junitAssertNoErrors()
	// -------------------------------------------------------------

	/**
	 * This implementation is NOT efficient because extra file I/O is done
	 * if default reporting is enabled. It also assumes the Client correctly invoked Validator.junitAssertNoErrors()
	 * 
	 * 1. Find the appropriate TestCase & write the file
	 * 2. Update the co-responding Suite file (may be extra writes)
	 * 3. Update the co-responding Package file (may be extra writes)
	 * 4. Update QAToolsOrg-detail.html
	 * 
	 * 5. Update the co-responding ErrorDetail file (may be extra writes)
	 * 6. Update QAToolsOrg-errorSummary.html
	 */
	public static void writeTestCaseReports(StackTraceElement testCallerDetails) {
		if (!QATools.isDefaultReportEnabled()) {
			return;
		}
		
		//
		//
		String[] parts = testCallerDetails.getClassName().split("\\.");
		//System.out.println("parts.length = " + parts.length);
		String suite = parts[parts.length-1];
		String pack = testCallerDetails.getClassName().replace("."+suite, "");
		String testcaseName = testCallerDetails.getMethodName();

		//
		//
		Package p = getPackage(pack);
		Suite s = getSuite(p, suite);
		TestCase tc = getTestCase(s, testcaseName);
		
		try {
			if (QATools.isErrorSummaryReportEnabled()) {
				for (Detail d : tc.details) {
					ErrorDetail ed = d.errorDetail;
					writeErrorDetail(ed);
					writeErrorSummaryReport();
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		try {
			if (QATools.isDetailReportEnabled()) {
				//write in this order because pass/fail stats are tabulated
				writeTestCase(tc);
				writeSuite(s);
				writePackage(p);
				writeDetailReport();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		
	}
	
	// -------------------------------------------------------------
	// ----- Writing one file at a time
	// -------------------------------------------------------------

	/**
	 * 
	 */
	private static List<String> getPageHeader(String msg) {
		List<String> list = new LinkedList<String>();
		list.add("<html><head><title>" + msg + "</title></head>");
		list.add("<body><h3>" + msg + "</h3>");
		return list;
	}
	
	/**
	 * 
	 */
	private static List<String> getPageFooter() {
		List<String> list = new LinkedList<String>();
		list.add("</body>");
		list.add("</html>");
		return list;
	}

	/**
	 * 
	 * @param lines
	 * @return
	 */
	private static List<String> formatLines(List<Line> lines) {
		if (lines==null || lines.size()==0) {
			return null;
		}
		
		List<String> out = new LinkedList<String>();
		out.add("<table border=2>");
		for (Line l : lines) {
			String color = YELLOW;
			if (l.passFlag == Boolean.TRUE) {
				color = "green";
			} else if (l.passFlag == Boolean.FALSE) {
				color = "red";
			}
			
			String tmp = "<tr bgcolor=\"" + color + "\">";
			if (0 < l.num) {
				tmp += "<td>" + l.num + "</td>";
			}
			if (l.col1 != null) {
				tmp += "<td>" + l.col1;
				if (l.col1Link != null) {
					tmp += " <a href=\"" + l.col1Link + "\">(more)</a>";
				}
				tmp += "</td>";
			}	//if l.col1

			if (l.col2 != null) {
				tmp += "<td>" + l.col2;
				if (l.col2Link != null) {
					tmp += " <a href=\"" + l.col2Link + "\">(more)</a>";
				}
				tmp += "</td>";
			}	//if l.col1
			
			tmp += "</tr>";
			out.add(tmp);
		}	//for Line
		out.add("</table>");
		return out;
	}
	
	/**
	 * Writes this package to disk
	 */
	private static void writePackage(Package pack) throws Exception {
		List<String> page = getPageHeader("Package: " + pack.packageName);
		page.add(notEfficientMsg2());
		pack.numWrites++;
		
		pack.total=0;
		pack.fail=0;
		pack.pass=0;
		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<pack.suites.size(); i++) {
			Suite s = pack.suites.get(i);
			Line line = new Line();
			line.num = i+1;
			line.col1 = "Suite: " + s.suiteName;
			line.col1Link = "../../" + s.htmlFile.getFullPath();
			line.col2 = "Total: " + s.total + ", Pass: " + s.pass + ", Fail: " + s.fail;
			line.passFlag = s.passFlag;
			lines.add(line);
			
			pack.total += s.total;
			pack.pass += s.pass;
			pack.fail += s.fail;
		}
		page.addAll(formatLines(lines));
		page.addAll(getPageFooter());
		LocalFileUtil.writeLines(pack.htmlFile, page);
	}

	/**
	 * Writes this Suite to disk
	 */
	private static void writeSuite(Suite suite) throws Exception {
		List<String> page = getPageHeader("Package: " + suite.pack.packageName + "<br>Suite: " + suite.suiteName);
		page.add(notEfficientMsg2());
		suite.numWrites++;

		//
		// it will only be invoked once if the user sets up @AfterClass to invoke the DetailReport
		if (1 < suite.numWrites) {
			notEfficientWarning = true;
		}

		
		suite.total = suite.testcases.size();
		suite.pass=0;
		suite.fail=0;
		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<suite.testcases.size(); i++) {
			TestCase tc = suite.testcases.get(i);
			Line line = new Line();
			line.num = i+1;
			line.col1 = "Testcase: " + tc.testcaseName;
			line.col1Link = "../../" + tc.htmlFile.getFullPath();
			line.col2 = "Total: " + tc.total + ", Pass: " + tc.pass + ", Fail: " + tc.fail;
			line.passFlag = tc.passFlag;
			lines.add(line);
			
			if (tc.passFlag) {
				suite.pass++;
			} else {
				suite.fail++;
			}
				
		}
		page.addAll(formatLines(lines));

		page.addAll(getPageFooter());
		LocalFileUtil.writeLines(suite.htmlFile, page);
	}
	
	/**
	 * Writes this TestCase to disk
	 */
	private static void writeTestCase(TestCase tc) throws Exception {
		List<String> page = getPageHeader("Package: " + tc.suite.pack.packageName + "<br>Suite: " + tc.suite.suiteName + "<br>TestCase: " + tc.testcaseName);
		page.add(notEfficientMsg2());
		tc.numWrites++;

		//
		// details
		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<tc.details.size(); i++) {
			Detail d = tc.details.get(i);
			Line line = new Line();
			line.num = i+1;
			line.col1 = d.msg;
			line.col1Link = null;
			line.passFlag = d.passFlag;
			lines.add(line);
		}
		page.addAll(formatLines(lines));

		//
		// debugs
		if (tc.debugs != null && 0<tc.debugs.size()) {
			lines = new LinkedList<Line>();
			for (int i=0; i<tc.debugs.size(); i++) {
				Line line = new Line();
				line.num = i+1;
				line.col1 = tc.debugs.get(i);
				line.col1Link = null;
				lines.add(line);
			}
			page.add("<h4>Debug information</h4>");
			page.addAll(formatLines(lines));
		}

		
		page.addAll(getPageFooter());
		LocalFileUtil.writeLines(tc.htmlFile, page);
	}
	
	/**
	 * Writes the main detail page to disk
	 */
	private static void writeDetailReport() throws Exception {
		List<String> page = getPageHeader("Detail Report");
		page.add(notEfficientMsg());

		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<packages.size(); i++) {
			Package p = packages.get(i);
			       
			Line line = new Line();
			line.num = i+1;
			line.col1 = "Package: " + p.packageName;
			line.col1Link = "../" + p.htmlFile.getFullPath();
			line.col2 = "Total: " + p.total + ", Pass: " + p.pass + ", Fail: " + p.fail;
			line.passFlag = p.passFlag;
			lines.add(line);
		}
		page.addAll(formatLines(lines));
		page.addAll(getPageFooter());
		LocalFile lf = new LocalFile("QATools.org-detailReport.html");
		LocalFileUtil.writeLines(lf, page);
	}

	/**
	 * Writes this error detail to disk
	 */
	private static void writeErrorDetail(ErrorDetail errorDetail) throws Exception {
		if (errorDetail == null) {
			//System.out.println("writeErrorDetail - null");
			return;
		}
		errorDetail.numWrites++;
		
		List<String> page = getPageHeader("ErrorDetail: " + errorDetail.callingClass + "::" + errorDetail.callingMethod + "::" + errorDetail.callingLine);
		page.add(notEfficientMsg2());

		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<errorDetail.details.size(); i++) {
			Detail d = errorDetail.details.get(i);
			String tc = d.testcase.testcaseName;
			String suite = d.testcase.suite.suiteName;
			String pack = d.testcase.suite.pack.packageName;
			
			Line line = new Line();
			line.num = i+1;
			line.col1 = pack + " :: " + suite + " :: " + tc;
			line.col2 = d.msg;
			line.col2Link = "../../" + d.testcase.htmlFile.getFullPath();
			line.passFlag = d.passFlag;
			lines.add(line);
		}
		page.addAll(formatLines(lines));

		
		page.addAll(getPageFooter());
		LocalFileUtil.writeLines(errorDetail.htmlFile, page);
	}
	
	/**
	 * Writes the main error summary page to disk
	 */
	private static void writeErrorSummaryReport() throws Exception {
		List<String> page = getPageHeader("Error Summary Report");
		page.add(notEfficientMsg());
		
		Collections.sort(errorDetails);
		List<Line> lines = new LinkedList<Line>();
		for (int i=0; i<errorDetails.size(); i++) {
			ErrorDetail ed = errorDetails.get(i);
			if (1 < ed.numWrites) {
				notEfficientWarning = true;
			}
			Line line = new Line();
			line.num = 0;
			line.col1 = ed.details.size() + " instances of ";
			line.col2 = ed.callingClass + "::" + ed.callingMethod + "::" + ed.callingLine;
			line.col2Link = "../" + ed.htmlFile.getFullPath();
			line.passFlag = Boolean.FALSE;
			lines.add(line);
		}
		List<String> formattedLines = formatLines(lines);
		if (formattedLines!=null && 0<formattedLines.size()) {
			page.addAll(formattedLines);
		}
		page.addAll(getPageFooter());
		LocalFile lf = new LocalFile("QATools.org-errorSummaryReport.html");
		LocalFileUtil.writeLines(lf, page);
	}

	/**
	 * 
	 * @return
	 */
	private static String notEfficientMsg() {
		if (notEfficientWarning) {
			return "<h5>WARNING - Excessive File I/O - <a href=\"../" + notEfficientFile.getFullPath() + "\">Please Read</a></h5>";
		}
		return "";
	}
	/**
	 * 
	 * Extra ../ since these files are in a diff directory
	 */
	private static String notEfficientMsg2() {
		if (notEfficientWarning) {
			return "<h5>WARNING - Excessive File I/O - <a href=\"../../" + notEfficientFile.getFullPath() + "\">Please Read</a></h5>";
		}
		return "";
	}


	// -------------------------------------------------------------
	// ----- Generates / writes all files at once
	// ----- This is more efficient than the default : writeTestCaseReports()
	// -------------------------------------------------------------

	/**
	 * This is the efficient way to generate this report. THe client invokes this at the end and reduces the number of file I/O
	 */
	public static void generateErrorSummaryReport() {
		try {
			for (ErrorDetail ed : DetailReport.errorDetails) {
				DetailReport.writeErrorDetail(ed);
			}
			DetailReport.writeErrorSummaryReport();
		} catch (Exception e) {
			//
		}
	}
	
	/**
	 * This is the efficient way to generate this report. THe client invokes this at the end and reduces the number of file I/O
	 */
	public static void generateDetailReport() {
		try {
			// write the reports in this order because the stats are tabulated
			DetailReport.writeDetailReport();
			for (Package pack : DetailReport.packages) {
				for (Suite s : pack.suites) {
					for (TestCase tc : s.testcases) {
						DetailReport.writeTestCase(tc);
					}
					DetailReport.writeSuite(s);
				}	// for Suite
				DetailReport.writePackage(pack);
			}	// for Package
		} catch (Exception e) {
			//
		}
	}
	
	
	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------

	/**
	 * 
	 * @param testCallerDetails
	 * @param actualCallerDetails
	 * @param msg
	 */
	protected static void recordFail(StackTraceElement testCallerDetails, StackTraceElement actualCallerDetails, String msg) {
		recordResult(false, testCallerDetails, actualCallerDetails, msg);
	}

	/**
	 * 
	 * @param testCallerDetails
	 * @param actualCallerDetails
	 * @param msg
	 */
	protected static void recordPass(StackTraceElement testCallerDetails, StackTraceElement actualCallerDetails, String msg) {
		recordResult(true, testCallerDetails, actualCallerDetails, msg);
	}

	protected static void recordResult(boolean passFlag, StackTraceElement testCallerDetails, StackTraceElement actualCallerDetails, String msg) {
			//      System.out.println(testCallerDetails.getFileName() + "::" + testCallerDetails.getClassName() + "::" +
			//      testCallerDetails.getMethodName() + "::" + 
			// 		actualCallerDetails.getFileName() + "::" + 
			//		actualCallerDetails.getMethodName() +
			//      "::" + actualCallerDetails.getLineNumber() + ":: PASS " + msg);		String flag = (passFlag ? "Pass" : "Fail");
			
			//System.out.println("classname.full = " + testCallerDetails.getClassName());
			String[] parts = testCallerDetails.getClassName().split("\\.");
			//System.out.println("parts.length = " + parts.length);
			String suite = parts[parts.length-1];
			String pack = testCallerDetails.getClassName().replace("."+suite, "");
			String testCaseName = testCallerDetails.getMethodName();
			//System.out.println("pack = " + pack);	
			//System.out.println("suite = " + suite);	
			//System.out.println("testCaseName = " + testCaseName);	
			//System.out.println("msg = " + msg);	
			
			//
			// detail report
			Detail d = addDetail(passFlag, pack, suite, testCaseName, msg);

			//
			// keep track of error summaries
			if (!passFlag && QATools.isErrorSummaryReportEnabled()) {
				String actualCallingClass = actualCallerDetails.getClassName();
				String actualCallingMethod = actualCallerDetails.getMethodName();
				int actualCallingLine = actualCallerDetails.getLineNumber();
				ErrorDetail ed = getErrorDetail(actualCallingClass, actualCallingMethod, actualCallingLine);
				if (ed == null) {
					ed = new ErrorDetail(actualCallingClass, actualCallingMethod, actualCallingLine);
					//System.out.println("Adding error detail: " + actualCallingClass + ", " + actualCallingMethod + ", " + actualCallingLine);
					errorDetails.add(ed);
				}
				ed.details.add(d);
				d.errorDetail = ed;
				//System.out.println("adding detail to errorDetail :: " + actualCallingClass + ", " + actualCallingMethod + ", " + actualCallingLine);
			}
			
			
		}

	/**
	 * 
	 * @param passFlag
	 * @param pack
	 * @param suite
	 * @param testcaseName
	 * @param detail
	 */
	protected static Detail addDetail(boolean passFlag, String pack, String suite, String testcaseName, String detail) {
		Detail d = new Detail(passFlag, detail);
		if (!QATools.isDetailReportEnabled()) {
			return d;	// return without adding detail to list
		}
		
		Package p = getPackage(pack);
		if (p == null) {
			p = new Package(pack);
			packages.add(p);
			//System.out.println("Adding package: " + pack);
		}
		
		Suite s = getSuite(p, suite);
		if (s == null) {
			s = new Suite(suite);
			s.pack = p;
			p.suites.add(s);
			//System.out.println("Adding suite: " + suite);
		}
		
		TestCase tc = getTestCase(s, testcaseName);
		if (tc == null) {
			tc = new TestCase(testcaseName);
			tc.suite = s;
			s.testcases.add(tc);
//			System.out.println("Adding testcase: " + testcaseName);
		}
		
		d.testcase = tc;
		tc.details.add(d);
//		System.out.println("Adding detail: " + detail);
		tc.total++;		
		if (!passFlag) {
			p.passFlag = false;
			s.passFlag = false;
			tc.passFlag = false;
			tc.fail++;
		} else {
			tc.pass++;
		}
		return d;
	}
	
	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------
	
	protected static Package getPackage(String pack) {
		Package p = null;
		for (Package pp : packages) {
			if (pp.packageName.equals(pack)) {
				p = pp;
//				System.out.println("Found package: " + pack);
				break;
			}
		}
		return p;
	}

	protected static Suite getSuite(Package pack, String suite) {
		Suite p = null;
		for (Suite pp : pack.suites) {
			if (pp.suiteName.equals(suite)) {
				p = pp;
//				System.out.println("Found suite: " + suite);
				break;
			}
		}
		return p;
	}

	protected static TestCase getTestCase(Suite suite, String testcaseName) {
		TestCase p = null;
		for (TestCase pp : suite.testcases) {
			if (pp.testcaseName.equals(testcaseName)) {
				p = pp;
//				System.out.println("Found testcase: " + testcaseName);
				break;
			}
		}
		return p;
	}
	
	protected static ErrorDetail getErrorDetail(String actualCallingClass, String actualCallingMethod, int actualCallingLine) {
		ErrorDetail ed = null;
		for (ErrorDetail e : errorDetails) {
			if (e.callingClass.equals(actualCallingClass)
					&& e.callingMethod.equals(actualCallingMethod)
					&& e.callingLine == actualCallingLine
					) {
				ed = e;
				break;
			}
		}
		return ed;
	}

	// -------------------------------------------------------------
	// -----
	// -------------------------------------------------------------

	
	/**
	 * 
	 * @author hockeyeh
	 *
	 */
	private static class Package {
		public String packageName;
		public List<Suite> suites = new LinkedList<Suite>();
		public Boolean passFlag = Boolean.TRUE;
		public int total=0;
		public int pass=0;
		public int fail=0;
		public LocalFile htmlFile = new LocalFile(qatoolsOrgDir, "p-" + LongUtil.getUniqueLong() + ".html");
		public int numWrites = 0;
		public Package(String name) { packageName = name; }
		
	}
	
	private static class Suite {
		public String suiteName;
		public List<TestCase> testcases = new LinkedList<TestCase>();
		public Boolean passFlag = Boolean.TRUE;
		public int total=0;
		public int pass=0;
		public int fail=0;
		public Package pack;
		public LocalFile htmlFile = new LocalFile(qatoolsOrgDir, "s-" + LongUtil.getUniqueLong() + ".html");
		public int numWrites = 0;
		public Suite(String name) { suiteName = name; }
	}
	
	private static class TestCase {
		public String testcaseName;
		public List<Detail> details = new LinkedList<Detail>();
		public Boolean passFlag = Boolean.TRUE;
		public int total=0;
		public int pass=0;
		public int fail=0;
		public Suite suite;
		public LocalFile htmlFile = new LocalFile(qatoolsOrgDir, "tc-" + LongUtil.getUniqueLong() + ".html");
		public int numWrites = 0;
		public List<String> debugs;
		public TestCase(String name) { testcaseName = name; }
	}
	
	private static class Detail {
		public String msg;
		public Boolean passFlag = Boolean.TRUE;
		public TestCase testcase;
		public ErrorDetail errorDetail;
		public Detail(boolean passFlag, String detail) {msg = detail; this.passFlag = passFlag; }
	}
	protected static class ErrorDetail implements Comparable {
		public String callingClass;
		public String callingMethod;
		public long callingLine;
		public List<Detail> details = new LinkedList<Detail>();
		public LocalFile htmlFile = new LocalFile(qatoolsOrgDir, "ed-" + LongUtil.getUniqueLong() + ".html");
		public int numWrites = 0;
		public ErrorDetail(String callingClass, String callingMethod, long callingLine) { this.callingClass = callingClass; this.callingMethod = callingMethod; this.callingLine = callingLine; }
		
		public int compareTo(Object arg0) {
			return ((ErrorDetail) arg0).details.size() - this.details.size();
		}
	}
	
	private static class Line {
		public Boolean passFlag = null;	//null = Yellow Debug
		public int num;
		public String col1 = null;
		public String col2 = null;
		public String col1Link = null;
		public String col2Link = null;		
	}
	
}



